/*
 **************************************************************************
 *		SWI-Prolog ActiveX Extensions.
 *
 *	This software is provided in source and binary form free of charge to 
 *  users of SWI-Prolog.  No warranty of any kind is expressed or implied;
 *  in particular, this is entirely unsupported software.
 *
 *  Documentation for this product is available from the SWI-Prolog 
 *  web site (http://www.swi.psy.uva.nl/projects/SWI-Prolog/).
 *
 *	Version 1.0.1.
 *
 *	October 1, 1999
 *
 **************************************************************************
 */

//
//	ZSTR.CPP
//

#include <iostream>
#include <fstream>
#include <stdarg.h>
#include <stdio.h>
#include "zstr.h"

int _t_vsntprintf(char * pch, size_t cch, const char * pszFormat, va_list argList)
{
	return _vsnprintf( pch, cch, pszFormat, argList );
}

int _t_vsntprintf(wchar_t * pch, size_t cch, const wchar_t * pszFormat, va_list argList)
{
	return _vsnwprintf( pch, cch, pszFormat, argList );
}

void _t_assign ( const char * pasz, T_ZSTR<char> & zstr )
{
	zstr.Reset();
	zstr.T_ZSTR<char>::t_string::operator=(pasz);
}


void _t_assign ( const wchar_t * pwsz, T_ZSTR<wchar_t> & zstr )
{
	zstr.Reset();
	zstr.T_ZSTR<wchar_t>::t_string::operator=(pwsz);
}

void _t_assign ( const wchar_t * pwsz, T_ZSTR<char> & zstr )
{
	int cch = ::wcstombs( NULL, pwsz, INT_MAX );
	if (cch < 0)
		_THROW1(runtime_error("implict conversion of wide chars failed (1)"));
	int cb =  (cch + 1) * sizeof(char);
	char * psz = (char *) ::_alloca( cb);
	cch = ::wcstombs( psz, pwsz, cb );
	if ( cch >= 0 )
		psz[cch] = 0;

	zstr.Reset();
	zstr.T_ZSTR<char>::t_string::operator=(psz);
}

void _t_assign ( const char * pasz, T_ZSTR<wchar_t> & zstr )
{
	int cwch = ::mbstowcs(NULL, pasz, INT_MAX);
	if ( cwch < 0 )
		_THROW1(runtime_error("implict conversion of multibyte chars failed (1)"));
	
	wchar_t * pwsz = (wchar_t *) ::_alloca( (cwch + 1) * sizeof(wchar_t) );
	if ( ! pwsz )
		_THROW1(runtime_error("_alloca failed"));

	// The third argument here used to be INT_MAX, but when compiled for released
	// and optimized, mbstowcs would sometimes fail (return -1). This seems to fix
	// it.
	cwch = ::mbstowcs(pwsz, pasz, cwch);

	if ( cwch < 0 )
		_THROW1(runtime_error("implict conversion of multibyte chars failed (1)"));

	pwsz[cwch] = L'\0';

	zstr.Reset();
	zstr.T_ZSTR<wchar_t>::t_string::operator=(pwsz);
}



void T_ZSTR<char> :: ToLower ()
{
	char * tsz = const_cast<char*>(c_str());
	_strlwr( tsz );
}

void T_ZSTR<char> :: ToUpper ()
{
	char * tsz = const_cast<char*>(c_str());
	_strupr( tsz );
}


void T_ZSTR<wchar_t> :: ToLower ()
{
	wchar_t * tsz = const_cast<wchar_t*>(c_str());
	_wcslwr( tsz );
}

void T_ZSTR<wchar_t> :: ToUpper ()
{
	wchar_t * tsz = const_cast<wchar_t*>(c_str());
	_wcsupr( tsz );
}


int g_cbVsprintfMaxBufferSize = 1000000;

//
//	The unfortunate truth is that both _vsnprintf() and _vsnwprintf() return
//	a negative number in two distinct cases:  if the buffer is exhausted or if the
//	data is invalid.  For this reason, we have to allow for the latter possiblity
//	by placing a reasonable limit on the formatted result.
//
template <class _TCH>
static int t_VsprintZstr ( T_ZSTR<_TCH> & zs, const _TCH * tszcFmt, va_list valist )
{	
	//  Attempt to "sprintf" the buffer. If it fails, reallocate
	//   a larger buffer and try again.	
	const int cbMaxLimit = g_cbVsprintfMaxBufferSize;
	const int cbMin = 256;
	int cbMaxNew = 0;
	_TCH * psz = NULL;
	int cbOut = -1;
	do 
	{
		if ( cbMaxNew == 0 )
		{	//  First cycle
			cbMaxNew = cbMin;
		}
		else
		{	//  Allocate a new buffer, 150% as large as the last attempt
			cbMaxNew += cbMaxNew / 2;
			//  If overlimit, stop now
			assert( cbMaxNew < cbMaxLimit );
			if ( cbMaxNew >= cbMaxLimit )
				break;
		}
		psz = (_TCH *) _alloca( cbMaxNew * sizeof *psz );
		psz[0] = 0;
		cbOut = cbOut = _t_vsntprintf( psz, cbMaxNew-1, tszcFmt, valist );
	} 
	while (  cbOut < 0 ) ;

	// If successful, zero-terminate the formatted result and assign to the string
	if ( cbOut >= 0 )
	{
		psz[cbOut] = 0;
		zs = psz;
	}
	return cbOut;  // Return number of bytes formatted or negative number if error
}

int _t_VsprintZstr ( T_ZSTR<char> & zs, const char * tszcFmt, va_list valist )
{
	return t_VsprintZstr( zs, tszcFmt, valist );
}

int _t_VsprintZstr ( T_ZSTR<wchar_t> & zs, const wchar_t * tszcFmt, va_list valist )
{
	return t_VsprintZstr( zs, tszcFmt, valist );
}
